#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
    This implementation of Event is triggered when an BT instruction is being processed.
"""
from borb.pdf.canvas.event.event_listener import Event


class BeginTextEvent(Event):
    """
    This implementation of Event is triggered when an BT instruction is being processed.
    """

    #
    # CONSTRUCTOR
    #

    #
    # PRIVATE
    #

    #
    # PUBLIC
    #

    pass
